@@ -16,7 +16,6 @@ import os
 from unittest import mock
 
 import ddt
-import six
 
 from os_brick import exception
 from os_brick.initiator.connectors import base
@@ -259,7 +258,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         # Should work for string, unicode, and list
         wwns_luns = [
             ('1234567890123456', 1),
-            (six.text_type('1234567890123456'), 1),
+            (str('1234567890123456'), 1),
             (['1234567890123456', '1234567890123457'], 1),
             (['1234567890123456', '1234567890123457'], 1),
         ]
@@ -350,6 +349,49 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         self.assertEqual(expected_commands, self.cmds)
         return connection_info
 
+    @ddt.data(None, 'dm-18')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_mpath_device')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_sysfs_multipath_dm')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
+    @mock.patch.object(os.path, 'exists', return_value=True)
+    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
+    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
+    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
+    def test_connect_volume_multipath_found_dm(self, sysfs_dm_name,
+                                               check_valid_device_mock,
+                                               get_device_info_mock,
+                                               get_scsi_wwn_mock,
+                                               get_fc_hbas_info_mock,
+                                               get_fc_hbas_mock,
+                                               realpath_mock,
+                                               exists_mock,
+                                               wait_for_rw_mock,
+                                               find_mp_dev_mock,
+                                               sysfs_multipath_dm_mock,
+                                               wait_mpath_device_mock):
+
+        check_valid_device_mock.return_value = True
+        sysfs_multipath_dm_mock.return_value = sysfs_dm_name
+        self._test_connect_volume_multipath(get_device_info_mock,
+                                            get_scsi_wwn_mock,
+                                            get_fc_hbas_info_mock,
+                                            get_fc_hbas_mock,
+                                            realpath_mock,
+                                            exists_mock,
+                                            wait_for_rw_mock,
+                                            find_mp_dev_mock,
+                                            'rw',
+                                            True)
+        sysfs_multipath_dm_mock.assert_called_once()
+        if sysfs_dm_name:
+            wait_mpath_device_mock.assert_called_once_with(sysfs_dm_name)
+        else:
+            wait_mpath_device_mock.assert_not_called()
+
     @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
     @mock.patch.object(os.path, 'exists', return_value=True)
@@ -475,6 +517,9 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
             find_mp_dev_mock, 'rw', False)
 
         self.assertNotIn('multipathd_id', connection_info['data'])
+        # Ensure we don't call it with the real path
+        device_name = discover_mp_dev_mock.call_args[0][-1]
+        self.assertNotEqual(realpath_mock.return_value, device_name)
 
     @mock.patch.object(fibre_channel.FibreChannelConnector, 'get_volume_paths')
     def test_extend_volume_no_path(self, mock_volume_paths):
@@ -720,7 +765,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         # Check that we turn to lowercase target wwns
         key = 'target_wwns' if 'target_wwns' in target_info else 'target_wwn'
         wwns = target_info.get(key)
-        wwns = [wwns] if isinstance(wwns, six.string_types) else wwns
+        wwns = [wwns] if isinstance(wwns, str) else wwns
         wwns = [w.lower() for w in wwns]
         if wwns:
             self.assertEqual(wwns, conn['data'][key])
@@ -732,17 +777,22 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
 
     @ddt.data(('/dev/mapper/<WWN>', True),
               ('/dev/mapper/mpath0', True),
+              # Check real devices are properly detected as non multipaths
+              ('/dev/sda', False),
               ('/dev/disk/by-path/pci-1-fc-1-lun-1', False))
     @ddt.unpack
     @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.remove_scsi_device')
     @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.requires_flush')
-    @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.get_dev_path')
+    @mock.patch('os_brick.utils.get_dev_path')
     def test__remove_devices(self, path_used, was_multipath, get_dev_path_mock,
                              flush_mock, remove_mock):
+        exc = exception.ExceptionChainer()
         get_dev_path_mock.return_value = path_used
         self.connector._remove_devices(mock.sentinel.con_props,
                                        [{'device': '/dev/sda'}],
-                                       mock.sentinel.device_info)
+                                       mock.sentinel.device_info,
+                                       force=False, exc=exc)
+        self.assertFalse(bool(exc))
         get_dev_path_mock.assert_called_once_with(mock.sentinel.con_props,
                                                   mock.sentinel.device_info)
         flush_mock.assert_called_once_with('/dev/sda', path_used,
@@ -750,6 +800,39 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         remove_mock.assert_called_once_with('/dev/sda',
                                             flush=flush_mock.return_value)
 
+    @ddt.data(('/dev/mapper/<WWN>', True),
+              ('/dev/mapper/mpath0', True),
+              # Check real devices are properly detected as non multipaths
+              ('/dev/sda', False),
+              ('/dev/disk/by-path/pci-1-fc-1-lun-1', False))
+    @ddt.unpack
+    @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.remove_scsi_device')
+    @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.requires_flush')
+    @mock.patch('os_brick.utils.get_dev_path')
+    def test__remove_devices_fails(self, path_used, was_multipath,
+                                   get_dev_path_mock, flush_mock, remove_mock):
+        exc = exception.ExceptionChainer()
+        get_dev_path_mock.return_value = path_used
+        remove_mock.side_effect = Exception
+        self.connector._remove_devices(mock.sentinel.con_props,
+                                       [{'device': '/dev/sda'},
+                                        {'device': '/dev/sdb'}],
+                                       mock.sentinel.device_info,
+                                       force=True, exc=exc)
+        self.assertTrue(bool(exc))
+        get_dev_path_mock.assert_called_once_with(mock.sentinel.con_props,
+                                                  mock.sentinel.device_info)
+
+        expect_flush = [mock.call('/dev/sda', path_used, was_multipath),
+                        mock.call('/dev/sdb', path_used, was_multipath)]
+        self.assertEqual(len(expect_flush), flush_mock.call_count)
+        flush_mock.assert_has_calls(expect_flush)
+
+        expect_remove = [mock.call('/dev/sda', flush=flush_mock.return_value),
+                         mock.call('/dev/sdb', flush=flush_mock.return_value)]
+        self.assertEqual(len(expect_remove), remove_mock.call_count)
+        remove_mock.assert_has_calls(expect_remove)
+
     @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
     @mock.patch.object(os.path, 'exists', return_value=True)
@@ -811,3 +894,144 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
             'tee -a /sys/block/sdc/device/delete',
         ]
         self.assertEqual(expected_commands, self.cmds)
+
+    @ddt.data((False, Exception), (True, Exception), (False, None))
+    @ddt.unpack
+    @mock.patch.object(fibre_channel.FibreChannelConnector, '_remove_devices')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'multipath_del_map')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'flush_multipath_device')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
+    @mock.patch.object(os.path, 'exists', return_value=True)
+    @mock.patch.object(os.path, 'realpath')
+    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
+    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device_path')
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
+    def test_disconnect_volume_fails(self, ignore_exc, side_effect,
+                                     check_valid_device_mock,
+                                     find_mp_device_path_mock,
+                                     get_device_info_mock,
+                                     get_scsi_wwn_mock,
+                                     get_fc_hbas_info_mock,
+                                     get_fc_hbas_mock,
+                                     realpath_mock,
+                                     exists_mock,
+                                     wait_for_rw_mock,
+                                     flush_mock,
+                                     del_map_mock,
+                                     remove_mock):
+
+        flush_mock.side_effect = side_effect
+        del_map_mock.side_effect = side_effect
+
+        check_valid_device_mock.return_value = True
+        self.connector.use_multipath = True
+        get_fc_hbas_mock.side_effect = self.fake_get_fc_hbas
+        get_fc_hbas_info_mock.side_effect = self.fake_get_fc_hbas_info
+
+        wwn = '360002ac00000000000000b860000741c'
+        multipath_devname = f'/dev/disk/by-id/dm-uuid-mpath-{wwn}'
+        realpath_mock.return_value = '/dev/dm-1'
+        devices = {"device": multipath_devname,
+                   "id": wwn,
+                   "devices": [{'device': '/dev/sdb',
+                                'address': '1:0:0:1',
+                                'host': 1, 'channel': 0,
+                                'id': 0, 'lun': 1},
+                               {'device': '/dev/sdc',
+                                'address': '1:0:0:2',
+                                'host': 1, 'channel': 0,
+                                'id': 0, 'lun': 1}]}
+        get_device_info_mock.side_effect = devices['devices']
+        get_scsi_wwn_mock.return_value = wwn
+
+        location = '10.0.2.15:3260'
+        name = 'volume-00000001'
+        vol = {'id': 1, 'name': name}
+        initiator_wwn = ['1234567890123456', '1234567890123457']
+
+        find_mp_device_path_mock.return_value = '/dev/mapper/mpatha'
+
+        connection_info = self.fibrechan_connection(vol, location,
+                                                    initiator_wwn)
+        if side_effect and not ignore_exc:
+            self.assertRaises(exception.ExceptionChainer,
+                              self.connector.disconnect_volume,
+                              connection_info['data'], devices['devices'][0],
+                              force=True, ignore_errors=ignore_exc)
+        else:
+            self.connector.disconnect_volume(connection_info['data'],
+                                             devices['devices'][0],
+                                             force=True,
+                                             ignore_errors=ignore_exc)
+
+        flush_mock.assert_called_once_with(
+            find_mp_device_path_mock.return_value)
+
+        expected = [
+            mock.call(f'/dev/disk/by-path/pci-0000:05:00.2-fc-0x{wwn}-lun-1')
+            for wwn in initiator_wwn]
+        if side_effect:
+            del_map_mock.assert_called_once_with('dm-1')
+            expected.append(mock.call(find_mp_device_path_mock.return_value))
+        else:
+            del_map_mock.assert_not_called()
+
+        remove_mock.assert_called_once_with(connection_info['data'],
+                                            devices['devices'],
+                                            devices['devices'][0],
+                                            True, mock.ANY)
+        self.assertEqual(len(expected), realpath_mock.call_count)
+        realpath_mock.assert_has_calls(expected)
+
+    @ddt.data(None, mock.sentinel.addressing_mode)
+    @mock.patch.object(fibre_channel.FibreChannelConnector,
+                       '_get_host_devices')
+    @mock.patch.object(fibre_channel.FibreChannelConnector,
+                       '_get_possible_devices')
+    def test__get_possible_volume_paths(self, addressing_mode,
+                                        pos_devs_mock, host_devs_mock):
+        conn_props = {'targets': mock.sentinel.targets}
+        if addressing_mode:
+            conn_props['addressing_mode'] = addressing_mode
+        res = self.connector._get_possible_volume_paths(conn_props,
+                                                        mock.sentinel.hbas)
+        pos_devs_mock.assert_called_once_with(mock.sentinel.hbas,
+                                              mock.sentinel.targets,
+                                              addressing_mode)
+        host_devs_mock.assert_called_once_with(pos_devs_mock.return_value)
+        self.assertEqual(host_devs_mock.return_value, res)
+
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'lun_for_addressing')
+    @mock.patch.object(fibre_channel.FibreChannelConnector, '_get_pci_num')
+    def test__get_possible_devices(self, pci_mock, lun_mock):
+        pci_mock.side_effect = [
+            (mock.sentinel.platform1, mock.sentinel.pci_num1),
+            (mock.sentinel.platform2, mock.sentinel.pci_num2)]
+        hbas = [mock.sentinel.hba1, mock.sentinel.hba2]
+        lun_mock.side_effect = [mock.sentinel.lun1B, mock.sentinel.lun2B] * 2
+        targets = [('wwn1', mock.sentinel.lun1), ('wwn2', mock.sentinel.lun2)]
+
+        res = self.connector._get_possible_devices(
+            hbas, targets, mock.sentinel.addressing_mode)
+
+        self.assertEqual(2, pci_mock.call_count)
+        pci_mock.assert_has_calls([mock.call(mock.sentinel.hba1),
+                                   mock.call(mock.sentinel.hba2)])
+        self.assertEqual(4, lun_mock.call_count)
+        lun_mock.assert_has_calls([
+            mock.call(mock.sentinel.lun1, mock.sentinel.addressing_mode),
+            mock.call(mock.sentinel.lun2, mock.sentinel.addressing_mode)] * 2)
+        expected = [
+            (mock.sentinel.platform1, mock.sentinel.pci_num1, '0xwwn1',
+             mock.sentinel.lun1B),
+            (mock.sentinel.platform1, mock.sentinel.pci_num1, '0xwwn2',
+             mock.sentinel.lun2B),
+            (mock.sentinel.platform2, mock.sentinel.pci_num2, '0xwwn1',
+             mock.sentinel.lun1B),
+            (mock.sentinel.platform2, mock.sentinel.pci_num2, '0xwwn2',
+             mock.sentinel.lun2B),
+        ]
+        self.assertEqual(expected, res)
